/*
 * Copyright (C) 2005-2006 Andre Noll <maan@systemlinux.org>
 *
 *     This program is free software; you can redistribute it and/or modify
 *     it under the terms of the GNU General Public License as published by
 *     the Free Software Foundation; either version 2 of the License, or
 *     (at your option) any later version.
 *
 *     This program is distributed in the hope that it will be useful,
 *     but WITHOUT ANY WARRANTY; without even the implied warranty of
 *     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *     GNU General Public License for more details.
 *
 *     You should have received a copy of the GNU General Public License
 *     along with this program; if not, write to the Free Software
 *     Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111, USA.
 */

/** \file filter_chain.c common helper functions for filter input/output */

#include "para.h"
#include "list.h"
#include "filter.h"
#include "error.h"
#include "string.h"

DEFINE_FILTER_ARRAY(filters);

/**
 * call the init function of each supported filter
 *
 * \param all_filters the array of all supported filters
 * \sa filter::init
 */
void filter_init(struct filter *all_filters)
{
	struct filter *f;

	for (f = all_filters; f->name; f++)
		f->init(f);
}

/**
 * close and destroy a filter callback
 *
 * \param fcb the filter callback to close
 *
 * This removes \a fcb from the list of filter callbacks and calls
 * the close callback associated with \a fcb.
 */
static void close_filter_callback(struct filter_callback *fcb)
{
	PARA_NOTICE_LOG("closing filter_callback %p, data: %p\n", fcb, fcb->data);
	list_del(&fcb->node);
	fcb->close(fcb);
}

/**
 * close all callbacks of a filter node
 *
 * \param fn the filter node which contains the filter callbacks to be closed
 *
 * Call close_filter_callback() for each entry in the filter callback list
 * of \a fn.
 */
static void close_callbacks(struct filter_node *fn)
{
	struct filter_callback *fcb, *tmp;

	list_for_each_entry_safe(fcb, tmp, &fn->callbacks, node)
		close_filter_callback(fcb);
}

static void call_callbacks(struct filter_node *fn, char *inbuf, size_t inlen,
	char *outbuf, size_t outlen)
{
	struct filter_callback *fcb, *tmp;
	list_for_each_entry_safe(fcb, tmp, &fn->callbacks, node) {
		int ret;
		if (inlen && fcb->input_cb) {
			ret = fcb->input_cb(inbuf, inlen, fcb);
			if (ret < 0) {
				close_filter_callback(fcb);
				continue;
			}
		}
		if (!outlen || !fcb->output_cb)
			continue;
		ret = fcb->output_cb(outbuf, outlen, fcb);
		if (ret < 0)
			close_filter_callback(fcb);
	}
}

/**
 * call the convert function of each filter
 *
 * \param fci the filter chain containing the list of filter nodes.
 *
 * This is the core function of the filter subsystem. It loops over the list of
 * filter nodes determined by \a fci and calls the filter's convert function if
 * there is input available for the filter node in question. If the convert
 * function consumed some or all of its input data, all registered input
 * callbacks are called.  Similarly, if a convert function produced output, all
 * registerd output callbacks get called.
 *
 * \return The sum of output bytes produced by the convert functions on success,
 * negative return value on errors.
 *
 * \sa filter_node, filter#convert, filter_callback
 */
int filter_io(struct filter_chain_info *fci)
{
	struct filter_node *fn;
	char *ib;
	size_t *loaded;
	int conv, conv_total = 0;
again:
	ib = fci->inbuf;
	loaded = fci->in_loaded;
	conv = 0;
	list_for_each_entry(fn, &fci->filters, node) {
		int ret;
		if (*loaded && fn->loaded < fn->bufsize) {
			size_t old_fn_loaded = fn->loaded;
			PARA_DEBUG_LOG("fc %p loaded: %zd, calling %s convert\n", fci, *loaded, fn->filter->name);
			ret = fn->filter->convert(ib, *loaded, fn);
			if (ret < 0) {
				if (!fci->error)
					fci->error = -ret;
				return ret;
			}
			call_callbacks(fn, ib, ret, fn->buf + old_fn_loaded, fn->loaded - old_fn_loaded);
			*loaded -= ret;
			conv += ret;
			if (*loaded && ret) {
				PARA_DEBUG_LOG("moving %zd bytes in input buffer for %s filter\n",
					*loaded,  fn->filter->name);
				memmove(ib, ib + ret, *loaded);
			}
		}
		ib = fn->buf;
		loaded = &fn->loaded;
	}
//	PARA_DEBUG_LOG("loaded: %d\n", *loaded);
	conv_total += conv;
	if (conv)
		goto again;
	return conv_total;
}

/**
 * close all filter nodes and its callbacks
 *
 * \param fci the filter chain to close
 *
 * For each filter node determined by \a fci, call the close function of each
 * registered filter callback as well as the close function of the
 * corresponding filter.  Free all resources and destroy all callback lists and
 * the filter node list.
 *
 * \sa filter::close, filter_callback::close
 */
void close_filters(struct filter_chain_info *fci)
{
	struct filter_node *fn, *tmp;

	if (!fci)
		return;
	PARA_DEBUG_LOG("closing filter chain %p\n", fci);
	list_for_each_entry_safe(fn, tmp, &fci->filters, node) {
		PARA_NOTICE_LOG("closing %s filter callbacks (fci %p, fn %p)\n", fn->filter->name, fci, fn);
		close_callbacks(fn);
		PARA_NOTICE_LOG("closing %s filter (fci %p, fn %p)\n", fn->filter->name, fci, fn);
		fn->filter->close(fn);
		list_del(&fn->node);
		free(fn);
	}
}

/*
 * If the filter has a command line parser and options is not NULL, run it.
 * Returns filter_num on success, negative on errors
 */
static int parse_filter_args(int filter_num, char *options, void **conf)
{
	struct filter *f = &filters[filter_num];
	int i, argc = 2;
	char *dummy_args[] = {"", "", NULL};
	char **argv;

//	PARA_DEBUG_LOG("%s, options: %s, parser: %p\n", f->name,
//		options? options : "(none)", f->parse_config);
	if (!f->parse_config)
		return options? -E_BAD_FILTER_OPTIONS : filter_num;
	if (options) {
//		PARA_DEBUG_LOG("options: %s\n", options);
		argc = split_args(options, &argv, ' ');
//		PARA_DEBUG_LOG("argc = %d, argv[0]: %s\n", argc, argv[0]);
		for (i = argc; i >= 0; i--)
			argv[i + 1] = argv[i];
		argc += 2;
		*conf = f->parse_config(argc, argv);
	} else {
		/* is it OK to have no options? */
		*conf = f->parse_config(2, dummy_args);
	}
	return *conf? filter_num : -E_BAD_FILTER_OPTIONS;
}

/**
 * check the filter command line options
 *
 * \param fa the command line options
 * \param conf points to the filter configuration upon successful return
 *
 * Check if \a fa starts with a the name of a supported filter, followed by
 * a colon. If yes, call the command line parser of that filter.
 *
 * \return On success, the number of the filter is returned and \a conf
 * is initialized to point to the filter configuration determined by \a fa.
 * On errors, a negative value is returned.
 *
 * Note: If \a fa specifies a filter that has no command line parser success is
 * returned, and \a conf is initialized to \p NULL.
 *
 * \sa filter::parse_config
 */
int check_filter_arg(char *fa, void **conf)
{
	int j;

	*conf = NULL;
//	PARA_DEBUG_LOG("arg: %s\n", fa);
	for (j = 0; filters[j].name; j++) {
		const char *name = filters[j].name;
		size_t len = strlen(name);
		char c;
		if (strlen(fa) < len)
			continue;
		if (strncmp(name, fa, len))
			continue;
		c = fa[len];
		if (c && c != ' ')
			continue;
		if (c && !filters[j].parse_config)
			return -E_BAD_FILTER_OPTIONS;
		return parse_filter_args(j, c? fa + len + 1 : NULL, conf);
	}
	return -E_UNSUPPORTED_FILTER;
}

